"""Deprecated. Import from `griffe_pydantic` directly instead."""

import warnings
from typing import Any

from griffe_pydantic._internal import dynamic

# YORE: Bump 2: Remove file.


def __getattr__(name: str) -> Any:
    warnings.warn(
        "Importing from `griffe_pydantic.common` is deprecated. Import from `griffe_pydantic` directly instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    try:
        return getattr(dynamic, name)
    except AttributeError:
        return getattr(dynamic, name.removeprefix("_"))



from typing import Any
from pydantic import field_validator, model_validator, ConfigDict, BaseModel, Field


class ExampleModel(BaseModel):
    """An example model."""

    model_config = ConfigDict(frozen=False)

    field_without_default: str
    """Shows the *[Required]* marker in the signature."""

    field_plain_with_validator: int = 100
    """Show standard field with type annotation."""

    field_with_validator_and_alias: str = Field("FooBar", alias="BarFoo", validation_alias="BarFoo")
    """Shows corresponding validator with link/anchor."""

    field_with_constraints_and_description: int = Field(
        default=5, ge=0, le=100, description="Shows constraints within doc string."
    )

    @field_validator("field_with_validator_and_alias", "field_without_default", mode="before")
    @classmethod
    def check_max_length_ten(cls, v) -> str:
        """Show corresponding field with link/anchor."""
        if len(v) >= 10:
            raise ValueError("No more than 10 characters allowed")
        return v

    @model_validator(mode="before")
    @classmethod
    def lowercase_only(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Ensure that the field without a default is lowercase."""
        if isinstance(data.get("field_without_default"), str):
            data["field_without_default"] = data["field_without_default"].lower()
        return data



from __future__ import annotations

import ast
import sys
from typing import TYPE_CHECKING

from griffe import (
    Alias,
    Attribute,
    Class,
    Docstring,
    Expr,
    ExprCall,
    ExprKeyword,
    ExprName,
    Function,
    Kind,
    Module,
    dynamic_import,
    get_logger,
)

from griffe_pydantic._internal import common

if TYPE_CHECKING:
    from pathlib import Path


_logger = get_logger("griffe_pydantic")


def _inherits_pydantic(cls: Class) -> bool:
    """Tell whether a class inherits from a Pydantic model.

    Parameters:
        cls: A Griffe class.

    Returns:
        True/False.
    """
    for base in cls.bases:
        if isinstance(base, (ExprName, Expr)):
            base = base.canonical_path  # noqa: PLW2901
        if base in {"pydantic.BaseModel", "pydantic.main.BaseModel"}:
            return True

    return any(_inherits_pydantic(parent_class) for parent_class in cls.mro())


def _pydantic_validator(func: Function) -> ExprCall | None:
    """Return a function's `pydantic.field_validator` decorator if it exists.

    Parameters:
        func: A Griffe function.

    Returns:
        A decorator value (Griffe expression).
    """
    for decorator in func.decorators:
        if isinstance(decorator.value, ExprCall) and decorator.callable_path in {
            "pydantic.field_validator",
            "pydantic.model_validator",
        }:
            return decorator.value
    return None


def _process_attribute(attr: Attribute, cls: Class, *, processed: set[str]) -> None:
    """Handle Pydantic fields."""
    if attr.canonical_path in processed:
        return
    processed.add(attr.canonical_path)

    # Properties are not fields.
    if "property" in attr.labels:
        return

    # Presence of `class-attribute` label and absence of `instance-attribute` label
    # indicates that the attribute is annotated with `ClassVar` and should be ignored.
    if "class-attribute" in attr.labels and "instance-attribute" not in attr.labels:
        return

    kwargs = {}
    if isinstance(attr.value, ExprCall):
        kwargs = {
            argument.name: argument.value for argument in attr.value.arguments if isinstance(argument, ExprKeyword)
        }

        if (
            attr.value.function.canonical_path == "pydantic.Field"
            and len(attr.value.arguments) >= 1
            and not isinstance(attr.value.arguments[0], ExprKeyword)
            and attr.value.arguments[0] != "..."  # handle Field(...), i.e. no default
        ):
            kwargs["default"] = attr.value.arguments[0]

    elif attr.value is not None:
        kwargs["default"] = attr.value

    if attr.name == "model_config":
        config = {}
        for key, value in kwargs.items():
            if isinstance(value, str):
                try:
                    config[key] = ast.literal_eval(value)
                except ValueError:
                    config[key] = value
            else:
                config[key] = value
        cls.extra[common._self_namespace]["config"] = config
        return

    attr.labels.add("pydantic-field")
    attr.labels.discard("class-attribute")
    attr.labels.discard("instance-attribute")

    attr.value = kwargs.get("default")
    constraints = {kwarg: value for kwarg, value in kwargs.items() if kwarg not in {"default", "description"}}
    attr.extra[common._self_namespace]["constraints"] = constraints

    # Populate docstring from the field's `description` argument.
    if not attr.docstring and (docstring := kwargs.get("description")):
        try:
            attr.docstring = Docstring(ast.literal_eval(docstring), parent=attr)  # type: ignore[arg-type]
        except ValueError:
            _logger.debug(f"Could not parse description of field '{attr.path}' as literal, skipping")


def _process_function(func: Function, cls: Class, *, processed: set[str]) -> None:
    """Handle Pydantic field validators."""
    if func.canonical_path in processed:
        return
    processed.add(func.canonical_path)

    if isinstance(func, Alias):
        _logger.debug(f"Cannot yet process {func}")
        return

    if decorator := _pydantic_validator(func):
        fields = [ast.literal_eval(field) for field in decorator.arguments if isinstance(field, str)]
        common._process_function(func, cls, fields)


def _process_class(cls: Class, *, processed: set[str], schema: bool = False) -> None:
    """Finalize the Pydantic model data."""
    if cls.canonical_path in processed:
        return

    if not _inherits_pydantic(cls):
        return

    processed.add(cls.canonical_path)

    common._process_class(cls)

    if schema:
        import_path: Path | list[Path] = cls.package.filepath
        if isinstance(import_path, list):
            import_path = import_path[-1]
        if import_path.name == "__init__.py":
            import_path = import_path.parent
        import_path = import_path.parent
        try:
            true_class = dynamic_import(cls.path, import_paths=[import_path, *sys.path])
        except ImportError:
            _logger.debug(f"Could not import class {cls.path} for JSON schema")
            return
        cls.extra[common._self_namespace]["schema"] = common._json_schema(true_class)

    for member in cls.all_members.values():
        kind = member.kind
        if kind is Kind.ATTRIBUTE:
            _process_attribute(member, cls, processed=processed)  # type: ignore[arg-type]
        elif kind is Kind.FUNCTION:
            _process_function(member, cls, processed=processed)  # type: ignore[arg-type]
        elif kind is Kind.CLASS:
            _process_class(member, processed=processed, schema=schema)  # type: ignore[arg-type]


def _process_module(
    mod: Module,
    *,
    processed: set[str],
    schema: bool = False,
) -> None:
    """Handle Pydantic models in a module."""
    if mod.canonical_path in processed:
        return
    processed.add(mod.canonical_path)

    for cls in mod.classes.values():
        # Don't process aliases, real classes will be processed at some point anyway.
        if not cls.is_alias:
            _process_class(cls, processed=processed, schema=schema)

    for submodule in mod.modules.values():
        _process_module(submodule, processed=processed, schema=schema)



from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any

from griffe import (
    Class,
    Extension,
    Module,
    get_logger,
)

from griffe_pydantic._internal import dynamic, static

if TYPE_CHECKING:
    from griffe import ObjectNode


_logger = get_logger("griffe_pydantic")


class PydanticExtension(Extension):
    """Griffe extension for Pydantic."""

    def __init__(self, *, schema: bool = False) -> None:
        """Initialize the extension.

        Parameters:
            schema: Whether to compute and store the JSON schema of models.
        """
        super().__init__()
        self._schema = schema
        self._processed: set[str] = set()
        self._recorded: list[tuple[ObjectNode, Class]] = []

    def on_package_loaded(self, *, pkg: Module, **kwargs: Any) -> None:  # noqa: ARG002
        """Detect models once the whole package is loaded."""
        for node, cls in self._recorded:
            self._processed.add(cls.canonical_path)
            dynamic._process_class(node.obj, cls, processed=self._processed, schema=self._schema)
        static._process_module(pkg, processed=self._processed, schema=self._schema)

    def on_class_instance(self, *, node: ast.AST | ObjectNode, cls: Class, **kwargs: Any) -> None:  # noqa: ARG002
        """Detect and prepare Pydantic models."""
        # Prevent running during static analysis.
        if isinstance(node, ast.AST):
            return

        try:
            import pydantic  # noqa: PLC0415
        except ImportError:
            _logger.warning("could not import pydantic - models will not be detected")
            return

        if issubclass(node.obj, pydantic.BaseModel):
            self._recorded.append((node, cls))



"""griffe-pydantic package.

Griffe extension for Pydantic.
"""

from __future__ import annotations

from pathlib import Path

from griffe_pydantic._internal.extension import PydanticExtension


def get_templates_path() -> Path:
    """Return the templates directory path."""
    return Path(__file__).parent / "templates"


__all__: list[str] = ["PydanticExtension", "get_templates_path"]



from __future__ import annotations

from typing import Any, Callable

from griffe import (
    Attribute,
    Class,
    Docstring,
    Function,
    Kind,
    get_logger,
)
from pydantic.fields import FieldInfo

from griffe_pydantic._internal import common

_logger = get_logger("griffe_pydantic")


def _process_attribute(obj: Any, attr: Attribute, cls: Class, *, processed: set[str]) -> None:
    """Handle Pydantic fields."""
    if attr.canonical_path in processed:
        return
    processed.add(attr.canonical_path)
    if attr.name == "model_config":
        cls.extra[common._self_namespace]["config"] = obj
        return

    if not isinstance(obj, FieldInfo):
        return

    attr.labels = {"pydantic-field"}
    attr.value = obj.default
    constraints = {}
    for constraint in common._field_constraints:
        if (value := getattr(obj, constraint, None)) is not None:
            constraints[constraint] = value
    attr.extra[common._self_namespace]["constraints"] = constraints

    # Populate docstring from the field's `description` argument.
    if not attr.docstring and (docstring := obj.description):
        attr.docstring = Docstring(docstring, parent=attr)


def _process_function(obj: Callable, func: Function, cls: Class, *, processed: set[str]) -> None:
    """Handle Pydantic field validators."""
    if func.canonical_path in processed:
        return
    processed.add(func.canonical_path)
    if dec_info := getattr(obj, "decorator_info", None):
        common._process_function(func, cls, dec_info.fields)


def _process_class(obj: type, cls: Class, *, processed: set[str], schema: bool = False) -> None:
    """Detect and prepare Pydantic models."""
    common._process_class(cls)
    if schema:
        cls.extra[common._self_namespace]["schema"] = common._json_schema(obj)
    for member in cls.all_members.values():
        kind = member.kind
        if kind is Kind.ATTRIBUTE:
            _process_attribute(getattr(obj, member.name), member, cls, processed=processed)  # type: ignore[arg-type]
        elif kind is Kind.FUNCTION:
            _process_function(getattr(obj, member.name), member, cls, processed=processed)  # type: ignore[arg-type]



from __future__ import annotations

import json
from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from griffe import Attribute, Class, Function
    from pydantic import BaseModel

_self_namespace = "griffe_pydantic"
_mkdocstrings_namespace = "mkdocstrings"

_field_constraints = {
    "gt",
    "ge",
    "lt",
    "le",
    "multiple_of",
    "min_length",
    "max_length",
    "pattern",
    "allow_inf_nan",
    "max_digits",
    "decimal_place",
}


def _model_fields(cls: Class) -> dict[str, Attribute]:
    return {name: attr for name, attr in cls.all_members.items() if "pydantic-field" in attr.labels}  # type: ignore[misc]


def _model_validators(cls: Class) -> dict[str, Function]:
    return {name: func for name, func in cls.all_members.items() if "pydantic-validator" in func.labels}  # type: ignore[misc]


def _json_schema(model: type[BaseModel]) -> str:
    """Produce a model schema as JSON.

    Parameters:
        model: A Pydantic model.

    Returns:
        A schema as JSON.
    """
    return json.dumps(model.model_json_schema(), indent=2)


def _process_class(cls: Class) -> None:
    """Set metadata on a Pydantic model.

    Parameters:
        cls: The Griffe class representing the Pydantic model.
    """
    cls.labels.add("pydantic-model")
    cls.extra[_self_namespace]["fields"] = partial(_model_fields, cls)
    cls.extra[_self_namespace]["validators"] = partial(_model_validators, cls)
    cls.extra[_mkdocstrings_namespace]["template"] = "pydantic_model.html.jinja"


def _process_function(func: Function, cls: Class, fields: Sequence[str]) -> None:
    """Set metadata on a Pydantic validator.

    Parameters:
        cls: A Griffe function representing the Pydantic validator.
    """
    func.labels = {"pydantic-validator"}
    if fields and fields[0] == "*":
        targets = [member for member in cls.all_members.values() if "pydantic-field" in member.labels]
    else:
        targets = [cls.all_members[field] for field in fields]

    func.extra[_self_namespace].setdefault("targets", [])
    func.extra[_self_namespace]["targets"].extend(targets)
    for target in targets:
        target.extra[_self_namespace].setdefault("validators", [])
        target.extra[_self_namespace]["validators"].append(func)



"""Tests for the `extension` module."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest
from griffe import Extensions, temporary_inspected_package, temporary_visited_package

from griffe_pydantic._internal.extension import PydanticExtension

if TYPE_CHECKING:
    from mkdocstrings_handlers.python.handler import PythonHandler


code = """
    from pydantic import field_validator, ConfigDict, BaseModel, Field


    class ExampleParentModel(BaseModel):
        '''An example parent model.'''
        parent_field: str = Field(..., description="Parent field.")


    class ExampleModel(ExampleParentModel):
        '''An example child model.'''

        model_config = ConfigDict(frozen=False)

        field_without_default: str
        '''Shows the *[Required]* marker in the signature.'''

        field_plain_with_validator: int = 100
        '''Show standard field with type annotation.'''

        field_with_validator_and_alias: str = Field("FooBar", alias="BarFoo", validation_alias="BarFoo")
        '''Shows corresponding validator with link/anchor.'''

        field_with_constraints_and_description: int = Field(
            default=5, ge=0, le=100, description="Shows constraints within doc string."
        )

        @field_validator("field_with_validator_and_alias", "field_plain_with_validator", mode="before")
        @classmethod
        def check_max_length_ten(cls, v):
            '''Show corresponding field with link/anchor.'''
            if len(v) >= 10:
                raise ValueError("No more than 10 characters allowed")
            return v

        def regular_method(self):
            pass


    class RegularClass(object):
        regular_attr = 1
"""


@pytest.mark.parametrize("analysis", ["static", "dynamic"])
def test_extension(analysis: str) -> None:
    """Test the extension."""
    loader = {"static": temporary_visited_package, "dynamic": temporary_inspected_package}[analysis]
    with loader(
        "package",
        modules={"__init__.py": code},
        extensions=Extensions(PydanticExtension(schema=True)),
    ) as package:
        assert package

        assert "ExampleParentModel" in package.classes
        assert package.classes["ExampleParentModel"].labels == {"pydantic-model"}

        assert "ExampleModel" in package.classes
        assert package.classes["ExampleModel"].labels == {"pydantic-model"}

        config = package.classes["ExampleModel"].extra["griffe_pydantic"]["config"]
        assert config == {"frozen": False}

        schema = package.classes["ExampleModel"].extra["griffe_pydantic"]["schema"]
        assert schema.startswith('{\n  "description"')


def test_imported_models() -> None:
    """Test the extension with imported models."""
    with temporary_visited_package(
        "package",
        modules={
            "__init__.py": "from ._private import MyModel\n\n__all__ = ['MyModel']",
            "_private.py": "from pydantic import BaseModel\n\nclass MyModel(BaseModel):\n    field1: str\n    '''Some field.'''\n",
        },
        extensions=Extensions(PydanticExtension(schema=False)),
    ) as package:
        assert package["MyModel"].labels == {"pydantic-model"}
        assert package["MyModel.field1"].labels == {"pydantic-field"}


def test_rendering_model_config_using_configdict(python_handler: PythonHandler) -> None:
    """Test the extension with model config using ConfigDict."""
    code = """
    from pydantic import BaseModel, ConfigDict, Field

    class Model(BaseModel):
        usage: str | None = Field(
            None,
            description="Some description.",
            example="Some example.",
        )
        model_config = ConfigDict(
            json_schema_extra={
                "example": {
                    "usage": "Some usage.",
                    "limitations": "Some limitations.",
                    "billing": "Some value.",
                    "notice_period": "Some value.",
                }
            }
        )
    """
    with temporary_visited_package(
        "package",
        modules={"__init__.py": code},
        extensions=Extensions(PydanticExtension(schema=False)),
    ) as package:
        python_handler.render(package["Model"], python_handler.get_options({}))  # Assert no errors.


def test_not_crashing_on_dynamic_field_description(caplog: pytest.LogCaptureFixture) -> None:
    """Test the extension with dynamic field description."""
    code = """
    import pydantic

    desc = "xyz"

    class TestModel(pydantic.BaseModel):
        abc: str = pydantic.Field(description=desc)
    """
    with (
        caplog.at_level(logging.DEBUG),
        temporary_visited_package(
            "package",
            modules={"__init__.py": code},
            extensions=Extensions(PydanticExtension(schema=False)),
        ),
    ):
        assert any(
            record.levelname == "DEBUG" and "field 'package.TestModel.abc' as literal" in record.message
            for record in caplog.records
        )


def test_ignore_classvars() -> None:
    """Test the extension ignores class variables."""
    code = """
    from pydantic import BaseModel
    from typing import ClassVar

    class Model(BaseModel):
        field: str
        class_var: ClassVar[int] = 1
    """
    with temporary_visited_package(
        "package",
        modules={"__init__.py": code},
        extensions=Extensions(PydanticExtension(schema=False)),
    ) as package:
        assert "pydantic-field" not in package["Model.class_var"].labels
        assert "class-attribute" in package["Model.class_var"].labels


def test_wildcard_field_validator() -> None:
    """Test field validator that works on all fields."""
    code = """
    from pydantic import BaseModel, field_validator

    class Schema(BaseModel):
        a: int
        b: int

        @field_validator('*', mode='before')
        @classmethod
        def set_if_none(cls, v: Any, info):
            ...
    """
    with temporary_visited_package(
        "package",
        modules={"__init__.py": code},
        extensions=Extensions(PydanticExtension(schema=False)),
    ) as package:
        validator = package["Schema.set_if_none"]
        assert validator.labels == {"pydantic-validator"}
        assert validator in package["Schema.a"].extra["griffe_pydantic"]["validators"]
        assert validator in package["Schema.b"].extra["griffe_pydantic"]["validators"]


def test_ignoring_properties() -> None:
    """Properties are not fields and must be ignored."""
    code = """
    from pydantic import BaseModel, field

    class Base(BaseModel):
        @property
        def a(self) -> int:
            return 0

    class Model(Base):
        b: int = field(default=1)
    """
    with temporary_visited_package(
        "package",
        modules={"__init__.py": code},
        extensions=Extensions(PydanticExtension(schema=False)),
    ) as package:
        assert "pydantic-field" not in package["Model.a"].labels


def test_process_non_model_base_class_fields() -> None:
    """Fields in a non-model base class must be processed."""
    code = """
    from pydantic import BaseModel, field

    class A:
        a: int = 0

    class B(BaseModel, A):
        b: int = 1
    """
    with temporary_visited_package(
        "package",
        modules={"__init__.py": code},
        extensions=Extensions(PydanticExtension(schema=False)),
    ) as package:
        assert "pydantic-field" in package["B.a"].labels
